/*
    net_setup.c -- Setup.
    Copyright (C) 1998-2005 Ivo Timmermans,
                  2000-2009 Guus Sliepen <guus@tinc-vpn.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id$
*/

#include "system.h"

#include "splay_tree.h"
#include "cipher.h"
#include "conf.h"
#include "connection.h"
#include "control.h"
#include "device.h"
#include "digest.h"
#include "graph.h"
#include "logger.h"
#include "net.h"
#include "netutl.h"
#include "process.h"
#include "protocol.h"
#include "route.h"
#include "rsa.h"
#include "subnet.h"
#include "utils.h"
#include "xalloc.h"

char *myport;
static struct event device_ev;

bool read_rsa_public_key(connection_t *c) {
	FILE *fp;
	char *fname;
	char *n;
	bool result;

	cp();

	/* First, check for simple PublicKey statement */

	if(get_config_string(lookup_config(c->config_tree, "PublicKey"), &n)) {
		result = rsa_set_hex_public_key(&c->rsa, n, "FFFF");
		free(n);
		return result;
	}

	/* Else, check for PublicKeyFile statement and read it */

	if(!get_config_string(lookup_config(c->config_tree, "PublicKeyFile"), &fname))
		asprintf(&fname, "%s/hosts/%s", confbase, c->name);

	fp = fopen(fname, "r");

	if(!fp) {
		logger(LOG_ERR, _("Error reading RSA public key file `%s': %s"),
			   fname, strerror(errno));
		free(fname);
		return false;
	}

	result = rsa_read_pem_public_key(&c->rsa, fp);
	fclose(fp);

	if(!result) 
		logger(LOG_ERR, _("Reading RSA public key file `%s' failed: %s"), fname, strerror(errno));
	free(fname);
	return result;
}

bool read_rsa_private_key() {
	FILE *fp;
	char *fname;
	char *n, *d;
	bool result;

	cp();

	/* First, check for simple PrivateKey statement */

	if(get_config_string(lookup_config(config_tree, "PrivateKey"), &d)) {
		if(!get_config_string(lookup_config(myself->connection->config_tree, "PublicKey"), &n)) {
			logger(LOG_ERR, _("PrivateKey used but no PublicKey found!"));
			free(d);
			return false;
		}
		result = rsa_set_hex_private_key(&myself->connection->rsa, n, "FFFF", d);
		free(n);
		free(d);
		return true;
	}

	/* Else, check for PrivateKeyFile statement and read it */

	if(!get_config_string(lookup_config(config_tree, "PrivateKeyFile"), &fname))
		asprintf(&fname, "%s/rsa_key.priv", confbase);

	fp = fopen(fname, "r");

	if(!fp) {
		logger(LOG_ERR, _("Error reading RSA private key file `%s': %s"),
			   fname, strerror(errno));
		free(fname);
		return false;
	}

#if !defined(HAVE_MINGW) && !defined(HAVE_CYGWIN)
	struct stat s;

	if(fstat(fileno(fp), &s)) {
		logger(LOG_ERR, _("Could not stat RSA private key file `%s': %s'"), fname, strerror(errno));
		free(fname);
		return false;
	}

	if(s.st_mode & ~0100700)
		logger(LOG_WARNING, _("Warning: insecure file permissions for RSA private key file `%s'!"), fname);
#endif

	result = rsa_read_pem_private_key(&myself->connection->rsa, fp);
	fclose(fp);

	if(!result) 
		logger(LOG_ERR, _("Reading RSA private key file `%s' failed: %s"), fname, strerror(errno));
	free(fname);
	return result;
}

static struct event keyexpire_event;

static void keyexpire_handler(int fd, short events, void *data) {
	regenerate_key();
}

void regenerate_key() {
	ifdebug(STATUS) logger(LOG_INFO, _("Regenerating symmetric key"));

	if(!cipher_regenerate_key(&myself->cipher, true)) {
		logger(LOG_ERR, _("Error regenerating key!"));
		abort();
	}

	if(timeout_initialized(&keyexpire_event)) {
		event_del(&keyexpire_event);
		send_key_changed(broadcast, myself);
	} else {
		timeout_set(&keyexpire_event, keyexpire_handler, NULL);
	}

	event_add(&keyexpire_event, &(struct timeval){keylifetime, 0});
}

/*
  Configure node_t myself and set up the local sockets (listen only)
*/
bool setup_myself(void) {
	config_t *cfg;
	subnet_t *subnet;
	char *name, *hostname, *mode, *afname, *cipher, *digest;
	char *address = NULL;
	char *envp[5];
	struct addrinfo *ai, *aip, hint = {0};
	bool choice;
	int i, err;

	cp();

	myself = new_node();
	myself->connection = new_connection();
	init_configuration(&myself->connection->config_tree);

	asprintf(&myself->hostname, _("MYSELF"));
	asprintf(&myself->connection->hostname, _("MYSELF"));

	myself->connection->options = 0;
	myself->connection->protocol_version = PROT_CURRENT;

	if(!get_config_string(lookup_config(config_tree, "Name"), &name)) {	/* Not acceptable */
		logger(LOG_ERR, _("Name for tinc daemon required!"));
		return false;
	}

	if(!check_id(name)) {
		logger(LOG_ERR, _("Invalid name for myself!"));
		free(name);
		return false;
	}

	myself->name = name;
	myself->connection->name = xstrdup(name);

	if(!read_connection_config(myself->connection)) {
		logger(LOG_ERR, _("Cannot open host configuration file for myself!"));
		return false;
	}

	if(!read_rsa_private_key())
		return false;

	if(!get_config_string(lookup_config(myself->connection->config_tree, "Port"), &myport))
		asprintf(&myport, "655");

	/* Read in all the subnets specified in the host configuration file */

	cfg = lookup_config(myself->connection->config_tree, "Subnet");

	while(cfg) {
		if(!get_config_subnet(cfg, &subnet))
			return false;

		subnet_add(myself, subnet);

		cfg = lookup_config_next(myself->connection->config_tree, cfg);
	}

	/* Check some options */

	if(get_config_bool(lookup_config(config_tree, "IndirectData"), &choice) && choice)
		myself->options |= OPTION_INDIRECT;

	if(get_config_bool(lookup_config(config_tree, "TCPOnly"), &choice) && choice)
		myself->options |= OPTION_TCPONLY;

	if(get_config_bool(lookup_config(myself->connection->config_tree, "IndirectData"), &choice) && choice)
		myself->options |= OPTION_INDIRECT;

	if(get_config_bool(lookup_config(myself->connection->config_tree, "TCPOnly"), &choice) && choice)
		myself->options |= OPTION_TCPONLY;

	if(myself->options & OPTION_TCPONLY)
		myself->options |= OPTION_INDIRECT;

	get_config_bool(lookup_config(config_tree, "TunnelServer"), &tunnelserver);

	if(get_config_string(lookup_config(config_tree, "Mode"), &mode)) {
		if(!strcasecmp(mode, "router"))
			routing_mode = RMODE_ROUTER;
		else if(!strcasecmp(mode, "switch"))
			routing_mode = RMODE_SWITCH;
		else if(!strcasecmp(mode, "hub"))
			routing_mode = RMODE_HUB;
		else {
			logger(LOG_ERR, _("Invalid routing mode!"));
			return false;
		}
		free(mode);
	} else
		routing_mode = RMODE_ROUTER;

	if(routing_mode == RMODE_ROUTER)
		if(!get_config_bool(lookup_config(myself->connection->config_tree, "PMTUDiscovery"), &choice) || choice)
			myself->options |= OPTION_PMTU_DISCOVERY;

	get_config_bool(lookup_config(config_tree, "PriorityInheritance"), &priorityinheritance);

#if !defined(SOL_IP) || !defined(IP_TOS)
	if(priorityinheritance)
		logger(LOG_WARNING, _("PriorityInheritance not supported on this platform"));
#endif

	if(!get_config_int(lookup_config(config_tree, "MACExpire"), &macexpire))
		macexpire = 600;

	if(get_config_int(lookup_config(config_tree, "MaxTimeout"), &maxtimeout)) {
		if(maxtimeout <= 0) {
			logger(LOG_ERR, _("Bogus maximum timeout!"));
			return false;
		}
	} else
		maxtimeout = 900;

	if(get_config_string(lookup_config(config_tree, "AddressFamily"), &afname)) {
		if(!strcasecmp(afname, "IPv4"))
			addressfamily = AF_INET;
		else if(!strcasecmp(afname, "IPv6"))
			addressfamily = AF_INET6;
		else if(!strcasecmp(afname, "any"))
			addressfamily = AF_UNSPEC;
		else {
			logger(LOG_ERR, _("Invalid address family!"));
			return false;
		}
		free(afname);
	}

	get_config_bool(lookup_config(config_tree, "Hostnames"), &hostnames);

	/* Generate packet encryption key */

	if(!get_config_string(lookup_config(myself->connection->config_tree, "Cipher"), &cipher))
		cipher = xstrdup("blowfish");

	if(!cipher_open_by_name(&myself->cipher, cipher)) {
		logger(LOG_ERR, _("Unrecognized cipher type!"));
		return false;
	}

	if(!get_config_int(lookup_config(config_tree, "KeyExpire"), &keylifetime))
		keylifetime = 3600;

	regenerate_key();

	/* Check if we want to use message authentication codes... */

	if(!get_config_string(lookup_config(myself->connection->config_tree, "Digest"), &digest))
		digest = xstrdup("sha1");

	if(!digest_open_by_name(&myself->digest, digest)) {
		logger(LOG_ERR, _("Unrecognized digest type!"));
		return false;
	}

	if(!get_config_int(lookup_config(myself->connection->config_tree, "MACLength"), &myself->maclength))

	if(digest_active(&myself->digest)) {
		if(myself->maclength > digest_length(&myself->digest)) {
			logger(LOG_ERR, _("MAC length exceeds size of digest!"));
			return false;
		} else if(myself->maclength < 0) {
			logger(LOG_ERR, _("Bogus MAC length!"));
			return false;
		}
	}

	/* Compression */

	if(get_config_int(lookup_config(myself->connection->config_tree, "Compression"), &myself->compression)) {
		if(myself->compression < 0 || myself->compression > 11) {
			logger(LOG_ERR, _("Bogus compression level!"));
			return false;
		}
	} else
		myself->compression = 0;

	myself->connection->outcompression = 0;

	/* Done */

	myself->nexthop = myself;
	myself->via = myself;
	myself->status.reachable = true;
	node_add(myself);

	graph();

	/* Open device */

	if(!setup_device())
		return false;

	event_set(&device_ev, device_fd, EV_READ|EV_PERSIST, handle_device_data, NULL);

	if (event_add(&device_ev, NULL) < 0) {
		logger(LOG_ERR, _("event_add failed: %s"), strerror(errno));
		close_device();
		return false;
	}

	/* Run tinc-up script to further initialize the tap interface */
	asprintf(&envp[0], "NETNAME=%s", netname ? : "");
	asprintf(&envp[1], "DEVICE=%s", device ? : "");
	asprintf(&envp[2], "INTERFACE=%s", iface ? : "");
	asprintf(&envp[3], "NAME=%s", myself->name);
	envp[4] = NULL;

	execute_script("tinc-up", envp);

	for(i = 0; i < 5; i++)
		free(envp[i]);

	/* Run subnet-up scripts for our own subnets */

	subnet_update(myself, NULL, true);

	/* Open sockets */

	get_config_string(lookup_config(config_tree, "BindToAddress"), &address);

	hint.ai_family = addressfamily;
	hint.ai_socktype = SOCK_STREAM;
	hint.ai_protocol = IPPROTO_TCP;
	hint.ai_flags = AI_PASSIVE;

	err = getaddrinfo(address, myport, &hint, &ai);

	if(err || !ai) {
		logger(LOG_ERR, _("System call `%s' failed: %s"), "getaddrinfo",
			   gai_strerror(err));
		return false;
	}

	listen_sockets = 0;

	for(aip = ai; aip; aip = aip->ai_next) {
		listen_socket[listen_sockets].tcp =
			setup_listen_socket((sockaddr_t *) aip->ai_addr);

		if(listen_socket[listen_sockets].tcp < 0)
			continue;

		listen_socket[listen_sockets].udp =
			setup_vpn_in_socket((sockaddr_t *) aip->ai_addr);

		if(listen_socket[listen_sockets].udp < 0) {
			close(listen_socket[listen_sockets].tcp);
			continue;
		}

		event_set(&listen_socket[listen_sockets].ev_tcp,
				  listen_socket[listen_sockets].tcp,
				  EV_READ|EV_PERSIST,
				  handle_new_meta_connection, NULL);
		if(event_add(&listen_socket[listen_sockets].ev_tcp, NULL) < 0) {
			logger(LOG_EMERG, _("event_add failed: %s"), strerror(errno));
			abort();
		}

		event_set(&listen_socket[listen_sockets].ev_udp,
				  listen_socket[listen_sockets].udp,
				  EV_READ|EV_PERSIST,
				  handle_incoming_vpn_data, NULL);
		if(event_add(&listen_socket[listen_sockets].ev_udp, NULL) < 0) {
			logger(LOG_EMERG, _("event_add failed: %s"), strerror(errno));
			abort();
		}

		ifdebug(CONNECTIONS) {
			hostname = sockaddr2hostname((sockaddr_t *) aip->ai_addr);
			logger(LOG_NOTICE, _("Listening on %s"), hostname);
			free(hostname);
		}

		memcpy(&listen_socket[listen_sockets].sa, aip->ai_addr, aip->ai_addrlen);
		listen_sockets++;

		if(listen_sockets >= MAXSOCKETS) {
			logger(LOG_WARNING, _("Maximum of %d listening sockets reached"), MAXSOCKETS);
			break;
		}
	}

	freeaddrinfo(ai);

	if(listen_sockets)
		logger(LOG_NOTICE, _("Ready"));
	else {
		logger(LOG_ERR, _("Unable to create any listening socket!"));
		return false;
	}

	return true;
}

/*
  setup all initial network connections
*/
bool setup_network_connections(void) {
	cp();

	init_connections();
	init_subnets();
	init_nodes();
	init_edges();
	init_requests();

	if(get_config_int(lookup_config(config_tree, "PingInterval"), &pinginterval)) {
		if(pinginterval < 1) {
			pinginterval = 86400;
		}
	} else
		pinginterval = 60;

	if(!get_config_int(lookup_config(config_tree, "PingTimeout"), &pingtimeout))
		pingtimeout = 5;
	if(pingtimeout < 1 || pingtimeout > pinginterval)
		pingtimeout = pinginterval;

	if(!get_config_int(lookup_config(config_tree, "MaxOutputBufferSize"), &maxoutbufsize))
		maxoutbufsize = 10 * MTU;

	if(!setup_myself())
		return false;

	try_outgoing_connections();

	return true;
}

/*
  close all open network connections
*/
void close_network_connections(void) {
	splay_node_t *node, *next;
	connection_t *c;
	char *envp[5];
	int i;

	cp();

	for(node = connection_tree->head; node; node = next) {
		next = node->next;
		c = node->data;
		c->outgoing = false;
		terminate_connection(c, false);
	}

	list_delete_list(outgoing_list);

	if(myself && myself->connection) {
		subnet_update(myself, NULL, false);
		terminate_connection(myself->connection, false);
		free_connection(myself->connection);
	}

	for(i = 0; i < listen_sockets; i++) {
		event_del(&listen_socket[i].ev_tcp);
		event_del(&listen_socket[i].ev_udp);
		close(listen_socket[i].tcp);
		close(listen_socket[i].udp);
	}

	asprintf(&envp[0], "NETNAME=%s", netname ? : "");
	asprintf(&envp[1], "DEVICE=%s", device ? : "");
	asprintf(&envp[2], "INTERFACE=%s", iface ? : "");
	asprintf(&envp[3], "NAME=%s", myself->name);
	envp[4] = NULL;

	exit_requests();
	exit_edges();
	exit_subnets();
	exit_nodes();
	exit_connections();

	execute_script("tinc-down", envp);

	if(myport) free(myport);

	for(i = 0; i < 4; i++)
		free(envp[i]);

	close_device();

	return;
}
